################################################################
# Name:          fixSolid
# Version:       1.5
# Description:   Try to fix solids
# 				 fixSolid fix the following problems:
#                      * heal missing faces
#                      * remove single edges
#                      * remove separate faces 
#                      * remove edges between coplanar faces
#                      * remove innerfaces
#                      * remove all edges with only 1 face connected (RISKY FIX)
# Author:       Guy Wydouw (guy.wydouw@skynet.be)
# Usage:        1. Select solid
#               2. Right-Click Context-Menu. If the solid is broken the option 'FIX SOLID' 
#                  is present. If the option isn't present there are no problems with
#                  the solid.
#               3. Select 'FIX SOLID'
#               4. You can set two options:
#						a. Risky fix: if you choose a risky fix it's possible that a part 
#                          of the solid wil be deleted. This action can be undone with the 
#                          normal SU undo function. The risky fix try to fix all problems, 
#                          the normal fix only fixes problems with no risk. 
#                       b. Fix coplanar faces: If you choose no, the edges between coplanar
#                          will not be removed.
# History: 
# 1.0  (16/12/2010) - first version
# 1.0b (17/12/2010) - removed debugging lines
# 1.1  (17/12/2010) - healfaces submitted
# 1.2  (17/12/2010) - fix crasch on single face solid
#                   - new dialogbox for options
#                   - option for fixing coplaner faces
# 1.3  (19/12/2010) - build-in check on minimum entities of 9 entities to prevent crashes
# 1.3b (19/12/2010) - new checking, when only 2 entities remaining fixSolid undo his operation
#                     and shows a message that the solid can't be fixed 
# 1.4  (20/12/2010) - new function for checking if faces are coplanar 
#                   - instance variable used instead of global
#                   - check if solid contains at least 2 edges
# 1.5  (20/12/2010) - module FixSolids
#                   - replaced .typename by .is_a?
################################################################

require 'sketchup'

module FixSolids

	@numfaces =0
	@test = false
	@errors = 0
	@solidtest = false
	@risky = false
	@solidundo = false
	@coplan = false
	@errorsremaining = 0

	def self.countedges(solid)
		numEdge = 0
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				numEdge = numEdge + 1
			end
		end
		return numEdge
	end	
	
	def self.healfaces(solid)
		@numfaces = 0
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				if e.faces.size == 1 then
					num = e.find_faces
					@numfaces = @numfaces + num
				end
			end
		end
	end

	def self.removelines(solid)
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				@test = false
				faces = e.faces
				if faces.size ==  0 then
					if self.solidcheck(solid) == false
						then return
					end
					e.erase!
					if self.solidcheck(solid) == false
						then return
					end
					@errors=@errors+1
					@test = true
					@solidTest = false
				end
			end
		end
	end

	def self.removeseparateface(solid)
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				@test = false
				faces = e.faces
				if faces.size == 1 then
					face = faces[0]
					edges = face.edges
					t=true
					edges.each do |zijde|
						if zijde.faces.size!=1 then
							t=false
						end
					end
					if t== true then
						face.erase!
						@errors=@errors+1
						@test = true
						@solidTest = false
					end
				end
			end
		end	
	end

	def self.sameplane?(face1,face2)
		vertices1 = face1.vertices
		vertices2 = face2.vertices
		
		plane = Geom.fit_plane_to_points vertices1[0].position, vertices1[1].position, vertices1[2].position
		
		vertices2.each do |v|
			pt = v.position
			result = pt.on_plane? plane
			if result == false then
				return false
			end
		end
		return true
	end

	def self.removecoplanaredge(solid)
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				@test = false
				faces = e.faces
				if faces.size == 2 then
					if self.sameplane?(faces[0], faces[1]) == true then
						e.erase!
						@errors=@errors+1
						@test = true
						@solidTest = false
					end
				end
			end
		end	
	end

	def self.removeinnerface(solid)
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				@test = false
				faces = e.faces
				if faces.size == 3 then
					faces.each do |face|
						edges = face.edges
						t=true
						edges.each do |edge|
							if edge.faces.size!=3 then
								t=false
							end
						end
						if t== true then
							face.erase!
							@errors=@errors+1
							@test = true
							@solidTest = false
						end
					end
				end
			end
		end	
	end

	def self.remove1faceedges(solid)
		solid.entities.each do |e|
			if e.is_a?(Sketchup::Edge) then
				@test = false
				if e.faces.size == 1 then
					e.erase!
					@errors=@errors+1
					@test = true
					@solidTest = false
				end
			end
		end
	end

	def self.fixSolid(solid)
		@solidTest = false
		
		self.healfaces(solid)
		until @solidTest == true
			@solidTest = true
			
			# remove open lines
			@test = true
			until @test == false
				self.removelines(solid)
			end

			# remove separate faces 
			@test = true
			until @test == false
				self.removeseparateface(solid)
			end		

			# remove edge between coplanar faces
			@test = true
			if @coplan == true then
				until @test == false
					self.removecoplanaredge(solid)
				end
			end

			# remove inner faces
			@test = true
			until @test == false
				self.removeinnerface(solid)
			end
			
			# remove edge only connected to 1 face
			if @risky == true then
				@test = true
				until @test == false
					self.remove1faceedges(solid)
				end
			end
			if solid.entities.count < 9 then
				@solidTest = true
			end
		end
	end

	def self.solidcheck(solid)
		if solid.entities.count < 3 then
			@solidundo = true
			return false
		end
		return true
	end

	UI.add_context_menu_handler do | menu |
		model = Sketchup.active_model
		entities = model.entities
		selection = model.selection
		layers = model.layers

		if selection[0].is_a?(Sketchup::Group) then
			solid = selection[0]
			if solid.manifold? == false then
				menu.add_item("FIX SOLID") {
					if self.countedges(solid) > 2 then
						@errors = 0
						
						 prompts = ["Risky fix? ", "Fix coplanar faces? "]
						 defaults = ["No", "Yes"]
						 list = ["Yes|No", "Yes|No"]
						 input = UI.inputbox prompts, defaults, list, "fixSolid options"
						 
						if input[0] == "Yes" then
							@risky = true
						else
							@risky = false
						end
						
						if input[1] == "Yes" then
							@coplan = true
						else
							@coplan = false
						end
						
						Sketchup.active_model.start_operation("FixSolids", true)
					
						self.fixSolid(solid)
						Sketchup.active_model.commit_operation
						
						if @solidundo == true then
							Sketchup.undo
							UI.messagebox "fixSolid can't fix the solid!", MB_OK
						else
							if solid.manifold? == false then
								selection.clear
								@errorsremaining=0
								solid.entities.each do |s|
									if s.is_a?(Sketchup::Edge) then
										if s.faces.size != 2 then
											selection.add s
											@errorsremaining = @errorsremaining + 1
										end
									end
								end
								
								errortxt = "Missing faces healed: " + @numfaces.to_s + "\n" + "Errors fixed: " + @errors.to_s + "\n" + "Remaining errors: " + @errorsremaining.to_s + "\n\n"
								UI.messagebox errortxt + " Solid NOT error free, remaining errors are selected.", MB_OK
									
							else
								errortxt = "Missing faces healed: " + @numfaces.to_s + "\n" + "Errors fixed: " + @errors.to_s + "\n"
								UI.messagebox errortxt + "Solid FIXED!", MB_OK
							end
						end
					else
						UI.messagebox "selected group is not a solid. A solid needs at least 2 edges!", MB_OK
					end
					
				}
			end
		end
	end
end

